from typing import List, Tuple, Union
import numpy as np
from pymilvus import Collection, connections, CollectionSchema, FieldSchema, DataType
from annb.indexes import IndexUnderTest, IndexUnderTestFactory, MetricType


class MilvusIndexUnderTest(IndexUnderTest):
    def __init__(
        self, index_name: str, dimension: int, metric_type: MetricType, **kwargs
    ):
        super().__init__(index_name, dimension, metric_type, **kwargs)
        self.connect()
        self.collection = self.create_collection()
        self.search_param = self.get_search_param()
        self.count = 0

    def get_search_param(self) -> dict:
        metric_type_text = self.get_index_param()['metric_type']
        search_param = {
            'metric_type': metric_type_text,
            'params': {}
        }
        if 'IVF' in metric_type_text:
            search_param['params']['nprobe'] = int(self.kwargs.get("nprobe", 128))
        if 'HNWS' in metric_type_text:
            search_param['params']['ef'] = int(self.kwargs.get("ef", 128))
        return search_param

    def get_index_param(self) -> str:
        index = self.kwargs.get("index", "IVF_FLAT")
        use_gpu = str(self.kwargs.get("gpu", "no")).lower() in [
            "yes",
            "true",
            "1",
            "on",
        ]
        use_gpu = use_gpu or 'GPU' in index
        metric_type_text = 'L2'
        if self.metric_type == MetricType.INNER_PRODUCT:
            metric_type_text = 'IP'
        #
        nlist = self.kwargs.get("nlist", 128)
        params = {}
        index_type = 'IVF_FLAT'
        if index in ('FLAT', 'flat'):
            index_type = 'FLAT'
        elif index in ('IVF_FLAT', 'ivfflat', 'GPU_IVF_FLAT'):
            if use_gpu:
                index_type = 'GPU_IVF_FLAT'
            else:
                index_type = 'IVF_FLAT'
            params = {
                'nlist': nlist,
            }
        elif index in ('IVF_SQ8', 'ivfsq8', 'ivfsq'):
            index_type = 'IVF_SQ8'
            params = {
                'nlist': nlist,
            }
        elif index in ('IVF_PQ', 'ivfpq', 'GPU_IVF_PQ'):
            if use_gpu:
                index_type = 'GPU_IVF_PQ'
            else:
                index_type = 'IVF_PQ'
            params = {
                'nlist': nlist,
                'm': int(self.kwargs.get("m", 16)),
                'nbits': int(self.kwargs.get("nbits", 8)),
            }
        elif index in ('HNSW', 'hnsw'):
            index_type = 'HNSW'
            params = {
                'M': int(self.kwargs.get("M", 16)),
                'efConstruction': int(self.kwargs.get("efConstruction", 256)),
            }
        return {
            'index_type': index_type,
            'metric_type': metric_type_text,
            'params': params,
        }

    def connect(self):
        uri = self.kwargs.get("uri", "http://localhost:19530")
        token = self.kwargs.get("token", "")
        connections.connect(uri=uri, token=token)

    def create_collection(self) -> Union[Collection, None]:
        schema = CollectionSchema(
            fields=[
                FieldSchema(
                    name='id',
                    dtype=DataType.INT64,
                    is_primary=True,
                ),
                FieldSchema(
                    name='vector',
                    dtype=DataType.FLOAT_VECTOR,
                    dim=self.dimension,
                )
            ],
            description="annb benchmark collection",
        )
        try:
            Collection('annb_collection').drop()
        except:
            pass
        return Collection(
            name='annb_collection',
            schema=schema,
            using='default',
            consistent_level='Strong',
        )

    def train(self, data: np.ndarray) -> None:
        pass

    def add(self, data: np.ndarray) -> None:
        add_count = data.shape[0]
        step_size = 1000000 // self.dimension
        for i in range(0, add_count, step_size):
            step_data = data[i : i + step_size]
            step_count = step_data.shape[0]
            ids = list(range(self.count, self.count + step_count))
            self.collection.insert([ids, step_data.tolist()])
            self.count += step_count

    def warmup(self) -> None:
        index_params = self.get_index_param()
        self.collection.create_index(
            field_name='vector',
            index_params=index_params,
            index_name='annb_benchmark_index'
        )
        self.collection.load()
        for _ in range(3):
            random_data = np.random.rand(10, self.dimension).astype("float32")
            random_data /= np.linalg.norm(random_data, axis=1)[:, None]
            self.search(random_data, 10)

    def search(self, query: np.ndarray, k: int) -> Tuple[List[float], List[int]]:
        result = self.collection.search(
            data=query,
            anns_field='vector',
            param=self.search_param,
            limit=k,
            consistency_level='Strong'
        )
        distances = []
        ids = []
        for r in result:
            distances.append(r.distances)
            ids.append(r.ids)
        return distances, ids

    def update_search_args(self, **kwargs):
        if "nprobe" in kwargs:
            self.search_param['params']['nprobe'] = kwargs['nprobe']
        if "ef" in kwargs:
            self.search_param['params']['ef'] = kwargs['ef']

    def cleanup(self) -> None:
        self.collection = self.create_collection()
        self.count = 0


class MilvusIndexUnderTestFactory(IndexUnderTestFactory):
    def create(
        self, index_name: str, dimension: int, metric_type: MetricType, **kwargs
    ) -> MilvusIndexUnderTest:
        return MilvusIndexUnderTest(index_name, dimension, metric_type, **kwargs)


index_under_test_factory = MilvusIndexUnderTestFactory



from typing import List, Tuple, Union
import numpy as np
import faiss
from annb.indexes import IndexUnderTest, IndexUnderTestFactory, MetricType


class FaissIndexUnderTest(IndexUnderTest):
    def __init__(
        self, index_name: str, dimension: int, metric_type: MetricType, **kwargs
    ):
        super().__init__(index_name, dimension, metric_type, **kwargs)
        self.index = self.create_index()

    def create_index(self) -> Union[faiss.Index, None]:
        faiss_metric = faiss.METRIC_L2
        if self.metric_type == MetricType.INNER_PRODUCT:
            faiss_metric = faiss.METRIC_INNER_PRODUCT
        elif self.metric_type == MetricType.JACCARD:
            # need faiss 1.7.4 or later
            faiss_metric = faiss.METRIC_Jaccard
        is_binary = False
        using_gpu = str(self.kwargs.get("gpu", "no")).lower() in [
            "yes",
            "true",
            "1",
            "on",
        ]
        index_string = self.kwargs.get("index", "flat")
        index = None
        if index_string == "flat":
            if is_binary:
                index = faiss.IndexBinaryFlat(self.dimension, faiss_metric)
                self.log.info("create index IndexBinaryFlat(d=%d,%s)", self.dimension, str(faiss_metric))
            else:
                index = faiss.IndexFlat(self.dimension, faiss_metric)
                self.log.info("create index IndexFlat(d=%d,%s)", self.dimension, str(faiss_metric))
        elif index_string == "ivfflat":
            nlist = self.kwargs.get("nlist", 128)
            if is_binary:
                index = faiss.IndexBinaryFlat(self.dimension, faiss_metric)
                index = faiss.IndexBinaryIVF(index, self.dimension, nlist)
                self.log.info("create index IndexBinaryIVF(d=%d,nlist=%d,%s)", self.dimension, nlist, str(faiss_metric))
            else:
                quantizer = faiss.IndexFlat(self.dimension, faiss_metric)
                index = faiss.IndexIVFFlat(quantizer, self.dimension, nlist, faiss_metric)
                self.log.info("create index IndexIVFFlat(d=%d,nlist=%d,%s)", self.dimension, nlist, str(faiss_metric))
        elif index_string == "ivfpq":
            quantizer = faiss.IndexFlat(self.dimension, faiss_metric)
            nlist = self.kwargs.get("nlist", 128)
            m = self.kwargs.get("m", 8)
            nbits = self.kwargs.get("nbits", 8)
            index = faiss.IndexIVFPQ(quantizer, self.dimension, nlist, m, nbits)
            self.log.info("create index IndexIVFPQ(d=%d,nlist=%d,m=%d,nbits=%d)", self.dimension, nlist, m, nbits)
        elif index_string == "ivfsq":
            quantizer = faiss.IndexFlat(self.dimension, faiss_metric)
            nlist = self.kwargs.get("nlist", 128)
            index = faiss.IndexIVFScalarQuantizer(
                quantizer,
                self.dimension,
                nlist,
                faiss.ScalarQuantizer.QT_8bit,
            )
            self.log.info("create index IndexIVFScalarQuantizer(d=%d,nlist=%d,QT_8bit)", self.dimension, nlist)
        else:
            index = faiss.index_factory(self.dimension, index_string, faiss_metric)
            self.log.info("create index %s(d=%d,%s)", index_string, self.dimension,str(faiss_metric))

        self.log.debug(
            "use cpu index, use gpu: %s, faiss has gpu support: %s",
            using_gpu,
            self.support_gpu(),
        )
        if using_gpu and self.support_gpu():
            self.log.debug("copy index to gpu")
            res = faiss.StandardGpuResources()
            index = faiss.index_cpu_to_gpu(res, 0, index)
        return index

    @classmethod
    def support_gpu(cls) -> bool:
        return hasattr(faiss, "get_num_gpus") and faiss.get_num_gpus() > 0

    def train(self, data: np.ndarray) -> None:
        return self.index.train(data)

    def add(self, data: np.ndarray) -> None:
        count = data.shape[0]
        step_size = 10000
        for i in range(0, count, step_size):
            self.index.add(data[i : i + step_size])
        return

    def warmup(self) -> None:
        for _ in range(3):
            random_data = np.random.rand(10, self.dimension).astype("float32")
            random_data /= np.linalg.norm(random_data, axis=1)[:, None]
            self.search(random_data, 10)

    def search(self, query: np.ndarray, k: int) -> Tuple[List[float], List[int]]:
        return self.index.search(query, k)

    def update_search_args(self, **kwargs):
        if "nprobe" in kwargs:
            self.index.nprobe = int(kwargs["nprobe"])

    def cleanup(self) -> None:
        self.index.reset()


class FaissIndexUnderTestFactory(IndexUnderTestFactory):
    def create(
        self, index_name: str, dimension: int, metric_type: MetricType, **kwargs
    ) -> FaissIndexUnderTest:
        return FaissIndexUnderTest(index_name, dimension, metric_type, **kwargs)


index_under_test_factory = FaissIndexUnderTestFactory



from annb.config import load_configs

def test_load_configs_foo(tmpdir):
    content = """
default:
  result: test

runs:
  - name: test1
  - name: test2
"""
    with open(tmpdir.join("test.yaml"), "w") as f:
        f.write(content)
    runs = load_configs(tmpdir.join("test.yaml"))



import numpy as np

from .hdf5_dataset import AnnbHdf5Dataset
from ..indexes import MetricType


class RandomDataset(AnnbHdf5Dataset):
    def __init__(self, file: str, **kwargs):
        metric_type = MetricType.from_text(kwargs.get('metric', 'l2'))
        dims = int(kwargs.get('dimension', 256))
        count = int(kwargs.get('count', 1000000))
        try:
            hdfile = AnnbHdf5Dataset.load_hdf5(file)
            temp_dataset = AnnbHdf5Dataset(hdfile)
            if temp_dataset.metric_type != metric_type:
                hdfile.close()
                raise RuntimeError(
                    f'Metric type mismatch: {temp_dataset.metric_type} != {metric_type}'
                )
            if temp_dataset.dimension != dims:
                hdfile.close()
                raise RuntimeError(
                    f'Dimension mismatch: {temp_dataset.dimension} != {dims}'
                )
            if temp_dataset.count != count:
                hdfile.close()
                raise RuntimeError(f'Count mismatch: {temp_dataset.count} != {count}')
        except (FileNotFoundError, RuntimeError):
            metric_type = MetricType.from_text(kwargs.get('metric', 'l2'))
            dims = int(kwargs.get('dimension', 256))
            count = int(kwargs.get('count', 1000000))
            normalize = kwargs.get('normalize', False)
            data = self.generate_data(dims, count, normalize)
            ground_truth = kwargs.get('ground_truth', True)
            AnnbHdf5Dataset.create(
                file, metric_type, data, normalize=normalize, ground_truth=ground_truth
            )
            hdfile = AnnbHdf5Dataset.load_hdf5(file)
        super().__init__(hdfile, **kwargs)

    @classmethod
    def generate_data(cls, dimension, count, normalize) -> np.ndarray:
        """
        Generate random data.
        """
        data = np.random.random((count, dimension)).astype(np.float32)
        return data



from time import monotonic
from typing import Tuple
import numpy as np
from ..indexes import MetricType


def execution(stage, func, *args, **kwargs):
    """
    Measure execution time of the function.

    :param stage: Stage name
    :param func: Function to measure
    :param args: Function args
    :param kwargs: Function kwargs

    :return: Function result
    """
    # duration_stage = stage.split('#')[0]
    started = monotonic()
    res = func(*args, **kwargs)
    duration = monotonic() - started
    print('Stage: {}, duration: {:.3f}ms'.format(stage, duration * 1000.0))
    return res


def generate_groundtruth(query, data, metric_type) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate ground truth for dataset.

    :param query: Query data
    :param data: Dataset data
    :param metric_type: Metric type

    :return: Ground truth
    """
    query_max = min(16384, query.shape[0])
    k = 100
    try:
        from faiss import knn_gpu
    except ImportError:
        knn_gpu = None

    try:
        from faiss import StandardGpuResources
    except ImportError:
        StandardGpuResources = None  # noqa

    if knn_gpu and StandardGpuResources:
        from faiss import METRIC_L2, METRIC_INNER_PRODUCT

        metric = METRIC_L2 if metric_type == MetricType.L2 else METRIC_INNER_PRODUCT
        res = StandardGpuResources()
        return execution(
            'generate_groundtruth/knn_gpu',
            knn_gpu,
            res,
            query[:query_max],
            data,
            k,
            metric=metric,
        )
    try:
        from faiss import knn
    except ImportError:
        knn = None

    if knn:
        from faiss import METRIC_L2, METRIC_INNER_PRODUCT

        metric = METRIC_L2 if metric_type == MetricType.L2 else METRIC_INNER_PRODUCT
        return execution(
            'generate_groundtruth/knn', knn, query[:query_max], data, k, metric=metric
        )

    try:
        from sklearn.neighbors import NearestNeighbors
    except ImportError:
        NearestNeighbors = None  # noqa

    if NearestNeighbors:
        metric_text = 'l2' if metric_type == MetricType.L2 else 'cosine'
        nbs = NearestNeighbors(n_neighbors=k, metric=metric_text, n_jobs=-1).fit(data)
        return execution(
            'generate_groundtruth/sklearn', nbs.kneighbors, query[:query_max]
        )

    raise RuntimeError('No knn implementation found')



from .indexes import MetricType
from .dataset import AnnbHdf5Dataset, Hdf5Dataset, RandomDataset, BaseDataset

__version__ = "0.1.19"

# metric types
METRIC_TYPE_INNER_PRODUCT = MetricType.INNER_PRODUCT
METRIC_TYPE_L2 = MetricType.L2

__all__ = [
    "METRIC_TYPE_INNER_PRODUCT",
    "METRIC_TYPE_L2",
    "BaseDataset",
    "AnnbHdf5Dataset",
    "Hdf5Dataset",
    "RandomDataset",
]
